## Graphical backend ##

## TODO:
# - Windows escape sequence
# - Add background coloring

ESCAPE = "\033"
COLOR_NAMES = ["black", "red", "green", "yellow", "blue", "purple", "aqua", "gray"]

class Rgb:
    # r: int
    # g: int
    # b: int
    def __init__(self, r: int, g: int, b: int):
        self.r = r
        self.b = b
        self.g = g

    def __repr__(self):
        return f"Rgb({self.r}, {self.g}, {self.b})"

    def __eq__(self, other):
        return (self.r, self.g, self.b) == (other.r, other.g, other.b)

    def to_hex(self, short=False):
        if short:
            if any(c % 17 for c in (self.r, self.g, self.b)):
                raise ValueError("Can't convert to short")

            return f"#{self.r // 17:x}{self.g // 17:x}{self.b // 17:x}"

        return f"#{self.r:0>2x}{self.g:0>2x}{self.b:0>2x}"

    @classmethod
    def from_hex(cls, color: str):
        """
        Return an Rgb instance from hex string

        color  - HEX-encoded color string
        """

        if not cls.is_hex(color):
            raise ValueError(f"Invalid HEX syntax: {color}")

        color = color.strip("#")

        # NOTE: No `match` for better version compat
        if len(color) == 6: # Full HEX
            r, g, b = color[:2], color[2:4], color[4:]

        elif len(color) == 3: # Short HEX
            r, g, b = color[0]*2, color[1]*2, color[2]*2

        return Rgb(int(r, 16), int(g, 16), int(b, 16))

    @classmethod
    def is_hex(cls, color: str):
        """
        Check if a string is a valid HEX value
        """

        color = color.strip("#")

        return len(color) in (3, 6) and all(char.isdigit() or char in "abcdef" for char in color)

def colorize(s: str, color: Rgb):
    """
    Return a string with escape sequence coloring applied

    s      - String to colorize
    color  - Color in RGB (3-element collection)
    """

    return f"{ESCAPE}[38;2;{color.r};{color.g};{color.b}m{s}{ESCAPE}[0m"

def colorize_ansi(s: str, name: str, bright=False):
    """
    Return a terminal color by its name

    name    - Name of the color from everyday English
    bright  - Whether the bright version should be returned
    """

    color = COLOR_NAMES.index(name) + bright*8

    return f"{ESCAPE}[38;5;{color}m{s}{ESCAPE}[0m"



## Substitution functions ##
import re

from . import gfx

# Parse a rules string into a dict
def parse_rules(rules_str):
    rules_lst = rules_str.strip("{} ").split(",")

    sub_rules = {}

    for pair in rules_lst:
        pattern = pair.split(":")[0].strip()
        color = pair.split(":")[1].strip()

        sub_rules.update({pattern: color})

    return sub_rules

# Perform colorizing substitution
def substitute(s, sub_rules):
        result = s

        for pattern in sub_rules:
            for match in set(re.finditer(pattern, s)):
                    match_str = match.string[match.start():match.end()]
                    color = sub_rules[pattern]

                    if color in gfx.COLOR_NAMES:
                        result = result.replace(match_str, gfx.colorize_ansi(match_str, color))
                    elif gfx.Rgb.is_hex(color):
                        color_rgb = gfx.Rgb.from_hex(color)

                        result = result.replace(match_str, gfx.colorize(match_str, color_rgb))
                    else:
                        raise ValueError(f"Invalid color syntax: {color}")

        return result



#!/usr/bin/env python3

## Terminal program/interface to the engine ##

import argparse
import sys

from . import gfx
from . import sub

stdin_input = True

# Parse args
argparser = argparse.ArgumentParser(
    prog="termcolors",
    epilog="if input isn't specified, STDIN is used."
)

argparser.add_argument("-r", "--rule", help="substitution ruleset")
argparser.add_argument("-R", "--rule-file", metavar="FILE", help="file to read substitution ruleset from")
argparser.add_argument("mode")
argparser.add_argument("input", nargs="*")

args = vars(argparser.parse_args())

def main():
    if args["input"] != []:
        stdin_input = False

    if args["mode"] == "substitute"[:len(args["mode"])]:
        # Read sub rules
        rules_str = ""

        if args["rule"]:
            rules_str += args["rule"].strip("{} ")

        if args["rule_file"]:
            with open(args["rule_file"], "r") as file:
                contents = file.read()

            stripped = contents.strip('{} ')

            rules_str += f", {stripped}"

        # Parse sub rules
        sub_rules = sub.parse_rules(rules_str)

        # Apply substitution
        if stdin_input:
            while True:
                inp = ""

                while (stdin := sys.stdin.read(1)) != "\n":
                    inp += stdin

                if inp == "":
                    break

                print(sub.substitute(inp, sub_rules))
        else:
            inp = "".join(args["input"]).strip()

            print(sub.substitute(inp, sub_rules))

    if args["mode"] == "check"[:len(args["mode"])]:
        from tests import demo_colors # Yes, yes, PEP 8

        demo_colors()
